package com.guo.service.impl;

import com.alibaba.fastjson2.JSONObject;
import com.guo.common.constant.LoadConstants;
import com.guo.gateway.cache.*;
import com.guo.gateway.model.NodeInfo;
import com.guo.gateway.model.NodeJoin;
import com.guo.gateway.model.WorkGatewayNode;
import lombok.AllArgsConstructor;
import lombok.Data;
import lombok.NoArgsConstructor;
import lombok.ToString;
import lombok.extern.log4j.Log4j2;
import org.redisson.api.RLock;
import org.redisson.api.RedissonClient;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicInteger;

/**
 * @author gxb
 * @description 负载实现层
 * @date 2024-04-18 14:30
 */
@Service
@AllArgsConstructor
@Log4j2
public class GateWayLoadServicelmpl implements GateWayLoadService {

    /**
     * 网关负载节点缓存
     */
    private final LoadNodeCache loadNodeCache;

    /**
     * 网关负载序列
     */
    private final LoadSeriesCache loadSeriesCache;

    /**
     * 网关节点缓存
     */
    private final NodeCache nodeCache;

    /**
     * 网关节点连接数
     */
    private final NodeScoreCache nodeScoreCache;

    /**
     * 网关节点存储VIN信息
     */
    private final NodeSetVinCache nodeSetVinCache;

    /**
     * 网关车辆对应网关节点
     */
    private final VehicleLineNodeCache vehicleLineNodeCache;

    /**
     * 分布式锁
     */
    //private final RedissonClient redissonClient;


    /**
     * 获取负载节点
     *
     * @return 返回公网IP
     */
    @Override
    public String loadNode() {
        //去刷新负载
        refreshLoad();
        //获取自增序列值
        Long seriesLoad = loadSeriesCache.incrementAndGet();  //获取自增序列值
        Long seriesLoadIndex = seriesLoad % LoadConstants.NODE_LENGTH;
        //获取负载下标
        String loadNodeId = loadNodeCache.getFindByIndex(seriesLoadIndex);
        //通过获取节点ID
        NodeInfo nodeInfo = nodeCache.get(loadNodeId);
        //获取缓存内节点的公网/内网信息 返回公网IP
        return nodeInfo.getPublicIdAddress();
    }

    /**
     * 刷新负载  实现动态负载
     */
    @Override
    public void refreshLoad() {

        //分布式锁
       // RLock refreshLoadLock = redissonClient.getLock("refreshLoadLock");
       // try {
            // 尝试获取锁，最多等待10秒，持有锁60秒后自动释放
           // if (refreshLoadLock.tryLock(10, 60, TimeUnit.SECONDS)) {
                // 在锁内执行刷新负载的逻辑
                List<WorkGatewayNode> workGatewayNodes = nodeScoreCache.getNodeScore();

                //车辆上线总数量
                long vehicleMaxOnlineNUm = workGatewayNodes.size() * 80;

                //目前连接数
                long veicleOnlineNowNum = Long.valueOf(String.valueOf(workGatewayNodes.stream().mapToDouble(WorkGatewayNode::getWeight).sum()));

                //空余连接数
                long vehicleOnlineNum = vehicleMaxOnlineNUm - veicleOnlineNowNum;

                //转换
                List<WorkGatewayNode> workGatewayNodeWeight = workGatewayNodes.stream()
                        .map(workGatewayNode -> WorkGatewayNode.builder()
                                .nodeId(workGatewayNode.getNodeId())
                                .weight(Integer.parseInt(String.valueOf(vehicleOnlineNum / (80L - workGatewayNode.getWeight()))))
                                .build())
                        .toList();

                List<String> loadNodeList = new ArrayList<>();

                int count = workGatewayNodeWeight.stream().mapToInt(WorkGatewayNode::getWeight).sum();

                if (count < 100) {
                    List<WorkGatewayNode> list = workGatewayNodeWeight.stream().sorted((o1, o2) -> o2.getWeight() - o1.getWeight()).toList();

                    int countWeight = 0;
                    for (long i = count; i < 100; i++) {
                        WorkGatewayNode workGatewayNode = list.get(countWeight++ % list.size());
                        workGatewayNode.setWeight(workGatewayNode.getWeight() + 1);
                    }
                }

                whFor:
                while (true) {
                    for (WorkGatewayNode workGatewayNode : workGatewayNodeWeight) {
                        int weight = workGatewayNode.getWeight();
                        if (weight > 0) {
                            loadNodeList.add(
                                    workGatewayNode.getNodeId()
                            );
                            workGatewayNode.setWeight(weight - 1);

                        }

                    }
                    int sum = workGatewayNodeWeight.stream().
                            mapToInt(WorkGatewayNode::getWeight).sum();
                    if (sum <= 0) {
                        break whFor;
                    }
                }
                //重置
                loadSeriesCache.reset();
                //存入负载集合
                loadNodeCache.put(loadNodeList);
            }
        //} catch (InterruptedException e) {
          //  Thread.currentThread().interrupt();
            // 处理中断异常
        //} finally {
            // 释放锁
           // if (refreshLoadLock.isHeldByCurrentThread()) {
               // refreshLoadLock.unlock();
//}


}

/**
 * 统计
 */
//class StiNode {
//    // 使用 ConcurrentHashMap 保证线程安全
//    private static final Map<String, AtomicInteger> stiNodeMap = new ConcurrentHashMap<>();
//
//    public static void sti(String nodeId) {
//        // 使用 computeIfAbsent 方法确保原子性操作
//        stiNodeMap.computeIfAbsent(nodeId, key -> new AtomicInteger()).incrementAndGet();
//    }
//
//    public static Map<String, Long> show() {
//        Map<String, Long> resultMap = new HashMap<>();
//        stiNodeMap.forEach((key, val) -> {
//            resultMap.put(key, (long) val.get());
//            System.out.println(key + "--------------" + val.get());
//        });
//        return resultMap;
//    }
//}



